@@ -1,4 +1,4 @@
-/* Copyright (C) 1996-2008, 2009, 2010, 2011 Free Software Foundation, Inc.
+/* Copyright (C) 1996-2011, 2012 Free Software Foundation, Inc.
    This file is part of the GNU C Library.
    Contributed by Ulrich Drepper <drepper@redhat.com>, 1996.
 
@@ -250,7 +250,7 @@ print_version (FILE *stream, struct argp_state *state)
 Copyright (C) %s Free Software Foundation, Inc.\n\
 This is free software; see the source for copying conditions.  There is NO\n\
 warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\
-"), "2011");
+"), "2012");
   fprintf (stream, gettext ("Written by %s.\n"), "Ulrich Drepper");
 }
 
@@ -470,7 +470,6 @@ this is the first definition"));
 	  else if (strncmp (&this_line[1], "delset", 6) == 0)
 	    {
 	      int cnt = sizeof ("delset");
-	      size_t set_number;
 	      while (isspace (this_line[cnt]))
 		++cnt;
 
@@ -492,11 +491,8 @@ this is the first definition"));
 		    ++cnt;
 
 		  if (cnt == start)
-		    {
-		      error_at_line (0, 0, fname, start_line,
-				     gettext ("illegal set number"));
-		      set_number = 0;
-		    }
+		    error_at_line (0, 0, fname, start_line,
+				   gettext ("illegal set number"));
 		  else
 		    {
 		      const char *symbol;
@@ -509,7 +505,6 @@ this is the first definition"));
 		      /* We have a symbolic set name.  This name must
 			 appear somewhere else in the catalogs read so
 			 far.  */
-		      set_number = 0;
 		      for (runp = current->all_sets; runp != NULL;
 			   runp = runp->next)
 			{
@@ -660,12 +655,10 @@ this is the first definition"));
 	  else if (ident[0] != '\0')
 	    {
 	      struct message_list *runp;
-	      struct message_list *lastp;
 
 	      /* Test whether the symbolic name was not used for
 		 another message in this message set.  */
 	      runp = current->current_set->messages;
-	      lastp = NULL;
 	      while (runp != NULL)
 		if (runp->symbol != NULL && strcmp (ident, runp->symbol) == 0)
 		  break;
