@@ -1,5 +1,5 @@
 #! @BASH@
-# Copyright (C) 1999, 2001-2008, 2009, 2010, 2011 Free Software Foundation, Inc.
+# Copyright (C) 1999, 2001-2011, 2012 Free Software Foundation, Inc.
 # This file is part of the GNU C Library.
 # Contributed by Ulrich Drepper <drepper@gnu.org>, 1999.
 
@@ -30,13 +30,13 @@ do_usage() {
 
 # Refer to --help option.
 help_info() {
-  printf >&2 $"Try \`xtrace --help' for more information.\n"
+  printf >&2 $"Try \`%s --help' or \`%s --usage' for more information.\n" xtrace xtrace
   exit 1
 }
 
 # Message for missing argument.
 do_missing_arg() {
-  printf >&2 $"xtrace: option \`$1' requires an argument.\n"
+  printf >&2 $"%s: option '%s' requires an argument.\n" xtrace "$1"
   help_info
 }
 
@@ -66,7 +66,7 @@ do_version() {
   printf $"Copyright (C) %s Free Software Foundation, Inc.
 This is free software; see the source for copying conditions.  There is NO
 warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-" "2011"
+" "2012"
   printf $"Written by %s.
 " "Ulrich Drepper"
   exit 0
