@@ -1,5 +1,5 @@
 /* Read and display shared object profiling data.
-   Copyright (C) 1997-2008, 2009, 2010, 2011 Free Software Foundation, Inc.
+   Copyright (C) 1997-2011, 2012 Free Software Foundation, Inc.
    This file is part of the GNU C Library.
    Contributed by Ulrich Drepper <drepper@cygnus.com>, 1997.
 
@@ -391,7 +391,7 @@ Copyright (C) %s Free Software Foundation, Inc.\n\
 This is free software; see the source for copying conditions.  There is NO\n\
 warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\
 "),
-	   "2011");
+	   "2012");
   fprintf (stream, gettext ("Written by %s.\n"), "Ulrich Drepper");
 }
 
@@ -407,7 +407,6 @@ load_shobj (const char *name)
   ElfW(Addr) mapend = 0;
   const ElfW(Phdr) *ph;
   size_t textsize;
-  unsigned int log_hashfraction;
   ElfW(Ehdr) *ehdr;
   int fd;
   ElfW(Shdr) *shdr;
@@ -477,13 +476,6 @@ load_shobj (const char *name)
   textsize = result->highpc - result->lowpc;
   result->kcountsize = textsize / HISTFRACTION;
   result->hashfraction = HASHFRACTION;
-  if ((HASHFRACTION & (HASHFRACTION - 1)) == 0)
-    /* If HASHFRACTION is a power of two, mcount can use shifting
-       instead of integer division.  Precompute shift amount.  */
-    log_hashfraction = __builtin_ffs (result->hashfraction
-				      * sizeof (struct here_fromstruct)) - 1;
-  else
-    log_hashfraction = -1;
   if (do_test)
     printf ("hashfraction = %d\ndivider = %Zu\n",
 	    result->hashfraction,
@@ -607,10 +599,11 @@ load_shobj (const char *name)
       static const char procpath[] = "/proc/self/fd/%d";
       char origprocname[sizeof (procpath) + sizeof (int) * 3];
       snprintf (origprocname, sizeof (origprocname), procpath, fd);
-      char *origlink = (char *) alloca (PATH_MAX + 1);
-      origlink[PATH_MAX] = '\0';
-      if (readlink (origprocname, origlink, PATH_MAX) == -1)
+      char *origlink = (char *) alloca (PATH_MAX);
+      ssize_t n = readlink (origprocname, origlink, PATH_MAX - 1);
+      if (n == -1)
 	goto no_debuginfo;
+      origlink[n] = '\0';
 
       /* Try to find the actual file.  There are three places:
 	 1. the same directory the DSO is in
