@@ -1,4 +1,4 @@
-/* Copyright (C) 2002, 2003, 2005, 2007, 2009 Free Software Foundation, Inc.
+/* Copyright (C) 2002,2003,2005,2007,2009,2011 Free Software Foundation, Inc.
    This file is part of the GNU C Library.
    Contributed by Ulrich Drepper <drepper@redhat.com>, 2002.
 
@@ -74,6 +74,29 @@ static const char *locnames[] =
 /* Size of the reserved address space area.  */
 #define RESERVE_MMAP_SIZE	512 * 1024 * 1024
 
+/* To prepare for enlargements of the mmaped area reserve some address
+   space.  On some machines, being a file mapping rather than an anonymous
+   mapping affects the address selection.  So do this mapping from the
+   actual file, even though it's only a dummy to reserve address space.  */
+static void *
+prepare_address_space (int fd, size_t total, size_t *reserved, int *xflags)
+{
+  if (total < RESERVE_MMAP_SIZE)
+    {
+      void *p = mmap64 (NULL, RESERVE_MMAP_SIZE, PROT_NONE, MAP_SHARED, fd, 0);
+      if (p != MAP_FAILED)
+        {
+          *reserved = RESERVE_MMAP_SIZE;
+          *xflags = MAP_FIXED;
+          return p;
+        }
+    }
+
+  *reserved = total;
+  *xflags = 0;
+  return NULL;
+}
+
 
 static void
 create_archive (const char *archivefname, struct locarhandle *ah)
@@ -81,7 +104,6 @@ create_archive (const char *archivefname, struct locarhandle *ah)
   int fd;
   char fname[strlen (archivefname) + sizeof (".XXXXXX")];
   struct locarhead head;
-  void *p;
   size_t total;
 
   strcpy (stpcpy (fname, archivefname), ".XXXXXX");
@@ -129,19 +151,9 @@ create_archive (const char *archivefname, struct locarhandle *ah)
       error (EXIT_FAILURE, errval, _("cannot resize archive file"));
     }
 
-  /* To prepare for enlargements of the mmaped area reserve some
-     address space.  */
-  size_t reserved = RESERVE_MMAP_SIZE;
-  int xflags = 0;
-  if (total < reserved
-      && ((p = mmap64 (NULL, reserved, PROT_NONE, MAP_PRIVATE | MAP_ANON,
-		       -1, 0)) != MAP_FAILED))
-    xflags = MAP_FIXED;
-  else
-    {
-      p = NULL;
-      reserved = total;
-    }
+  size_t reserved;
+  int xflags;
+  void *p = prepare_address_space (fd, total, &reserved, &xflags);
 
   /* Map the header and all the administration data structures.  */
   p = mmap64 (p, total, PROT_READ | PROT_WRITE, MAP_SHARED | xflags, fd, 0);
@@ -297,10 +309,8 @@ enlarge_archive (struct locarhandle *ah, const struct locarhead *head)
   int fd;
   struct locarhead newhead;
   size_t total;
-  void *p;
   unsigned int cnt, loccnt;
   struct namehashent *oldnamehashtab;
-  struct locrecent *oldlocrectab;
   struct locarhandle new_ah;
   struct oldlocrecent *oldlocrecarray;
   size_t prefix_len = output_prefix ? strlen (output_prefix) : 0;
@@ -392,19 +402,9 @@ enlarge_archive (struct locarhandle *ah, const struct locarhead *head)
       error (EXIT_FAILURE, errval, _("cannot resize archive file"));
     }
 
-  /* To prepare for enlargements of the mmaped area reserve some
-     address space.  */
-  size_t reserved = RESERVE_MMAP_SIZE;
-  int xflags = 0;
-  if (total < reserved
-      && ((p = mmap64 (NULL, reserved, PROT_NONE, MAP_PRIVATE | MAP_ANON,
-		       -1, 0)) != MAP_FAILED))
-    xflags = MAP_FIXED;
-  else
-    {
-      p = NULL;
-      reserved = total;
-    }
+  size_t reserved;
+  int xflags;
+  void *p = prepare_address_space (fd, total, &reserved, &xflags);
 
   /* Map the header and all the administration data structures.  */
   p = mmap64 (p, total, PROT_READ | PROT_WRITE, MAP_SHARED | xflags, fd, 0);
@@ -432,8 +432,6 @@ enlarge_archive (struct locarhandle *ah, const struct locarhead *head)
      still referenced and transfer it into the new file.  */
   oldnamehashtab = (struct namehashent *) ((char *) ah->addr
 					   + head->namehash_offset);
-  oldlocrectab = (struct locrecent *) ((char *) ah->addr
-				       + head->locrectab_offset);
 
   /* Sort the old locrec table in order of data position.  */
   oldlocrecarray = alloca (sizeof (*oldlocrecarray) * head->namehash_size);
@@ -609,20 +607,9 @@ open_archive (struct locarhandle *ah, bool readonly)
   ah->fd = fd;
   ah->mmaped = st.st_size;
 
-  /* To prepare for enlargements of the mmaped area reserve some
-     address space.  */
-  size_t reserved = RESERVE_MMAP_SIZE;
-  int xflags = 0;
-  void *p;
-  if (st.st_size < reserved
-      && ((p = mmap64 (NULL, reserved, PROT_NONE, MAP_PRIVATE | MAP_ANON,
-		       -1, 0)) != MAP_FAILED))
-    xflags = MAP_FIXED;
-  else
-    {
-      p = NULL;
-      reserved = st.st_size;
-    }
+  size_t reserved;
+  int xflags;
+  void *p = prepare_address_space (fd, st.st_size, &reserved, &xflags);
 
   /* Map the entire file.  We might need to compare the category data
      in the file with the newly added data.  */
@@ -1080,6 +1067,8 @@ add_locale_to_archive (ah, name, data, replace)
   int mask = _nl_explode_name (strdupa (name),
 			       &language, &modifier, &territory,
 			       &codeset, &normalized_codeset);
+  if (mask == -1)
+    return -1;
 
   if (mask & XPG_NORM_CODESET)
     /* This name contains a codeset in unnormalized form.
@@ -1129,6 +1118,7 @@ add_locale_to_archive (ah, name, data, replace)
 
   /* Now read the locale.alias files looking for lines whose
      right hand side matches our name after normalization.  */
+  int result = 0;
   if (alias_file != NULL)
     {
       FILE *fp;
@@ -1208,6 +1198,11 @@ add_locale_to_archive (ah, name, data, replace)
 						     &rhs_territory,
 						     &rhs_codeset,
 						     &rhs_normalized_codeset);
+		    if (rhs_mask == -1)
+		      {
+			result = -1;
+			goto out;
+		      }
 		    if (!strcmp (language, rhs_language)
 			&& ((rhs_mask & XPG_CODESET)
 			    /* He has a codeset, it must match normalized.  */
@@ -1241,6 +1236,7 @@ add_locale_to_archive (ah, name, data, replace)
 	    }
 	}
 
+    out:
       fclose (fp);
     }
 
@@ -1249,7 +1245,7 @@ add_locale_to_archive (ah, name, data, replace)
   if (mask & XPG_NORM_CODESET)
     free ((char *) normalized_codeset);
 
-  return 0;
+  return result;
 }
 
 
