@@ -1,5 +1,5 @@
 /* Run-time dynamic linker data structures for loaded ELF shared objects.
-   Copyright (C) 1995-2009, 2010 Free Software Foundation, Inc.
+   Copyright (C) 1995-2009, 2010, 2011 Free Software Foundation, Inc.
    This file is part of the GNU C Library.
 
    The GNU C Library is free software; you can redistribute it and/or
@@ -396,7 +396,7 @@ struct rtld_global
 	uint32_t hashval;
 	const char *name;
 	const ElfW(Sym) *sym;
-	struct link_map *map;
+	const struct link_map *map;
       } *entries;
       size_t size;
       size_t n_elements;
@@ -564,9 +564,10 @@ struct rtld_global_ro
 #define DL_DEBUG_FILES      (1 << 6)
 #define DL_DEBUG_STATISTICS (1 << 7)
 #define DL_DEBUG_UNUSED	    (1 << 8)
+#define DL_DEBUG_SCOPES	    (1 << 9)
 /* These two are used only internally.  */
-#define DL_DEBUG_HELP       (1 << 9)
-#define DL_DEBUG_PRELINK    (1 << 10)
+#define DL_DEBUG_HELP       (1 << 10)
+#define DL_DEBUG_PRELINK    (1 << 11)
 
   /* OS version.  */
   EXTERN unsigned int _dl_osversion;
@@ -955,8 +956,8 @@ extern void _dl_init (struct link_map *main_map, int argc, char **argv,
 extern void _dl_fini (void) internal_function;
 
 /* Sort array MAPS according to dependencies of the contained objects.  */
-extern void _dl_sort_fini (struct link_map *l, struct link_map **maps,
-			   size_t nmaps, char *used, Lmid_t ns)
+extern void _dl_sort_fini (struct link_map **maps, size_t nmaps, char *used,
+			   Lmid_t ns)
      internal_function;
 
 /* The dynamic linker calls this function before and having changing
@@ -1112,6 +1113,9 @@ extern void *_dl_tls_get_addr_soft (struct link_map *l) attribute_hidden;
 extern int _dl_addr_inside_object (struct link_map *l, const ElfW(Addr) addr)
      internal_function attribute_hidden;
 
+/* Show show of an object.  */
+extern void _dl_show_scope (struct link_map *new, int from);
+
 __END_DECLS
 
 #endif /* ldsodefs.h */
